/*
 * growisofs 5.5 by Andy Polyakov <appro@fy.chalmers.se>.
 *
 * Use-it-on-your-own-risk, GPL bless...
 *
 * This front-end to mkisofs(8) was originally developed to facilitate
 * appending of data to ISO9660 volumes residing on random write access
 * DVD media such as DVD+RW, DVD-RAM, as well as plain files/iso images.
 * At later stages even support for multi-session recording to DVD
 * write-once media such as DVD+R and DVD-R was added.
 *
 * As for growing random access volumes. The idea is very simple. The
 * program appends new data as it was added to a multisession media and
 * then copies the new volume descriptor(s) to the beginning of media
 * thus effectively updating the root catalog reference...
 *
 * For further details see http://fy.chalmers.se/~appro/linux/DVD+RW/.
 *
 * Revision history:
 *
 * 1.1:
 * - flush cache before copying volume descriptors;
 * 2.0:
 * - support for /dev/raw*;
 * - support for set-root-uid operation (needed for /dev/raw*);
 * - support for first "session" burning (needed for "poor-man");
 * - "poor-man" support for those who don't want to recompile the
 *   kernel;
 * 2.1:
 * - mkisofs_pid typo;
 * 2.2:
 * - uninitialized in_device variable;
 * - -help option;
 * 3.0:
 * - support for DVD+R;
 * 3.1:
 * - -Z fails if a file system is present and stdin is not a tty;
 * 3.2:
 * - support for image burning (needed for DVD+R as you can't use dd);
 * 3.3:
 * - 'growisofs -Z /dev/scdN image.iso' is too confusing, implement
 *   'growisofs -Z /dev/scdN=image.iso' instead;
 * 4.0:
 * - transport C++-fication for better portability;
 * - support for -dvd-compat option (improved DVD+R/RW compatibility);
 * - -dvd-video implies -dvd-compat;
 * - support for SONY DRU-500A;
 * - progress indicator for -Z /dev/scdN=image.iso;
 * - agressive -poor-man-ing;
 * 4.1:
 * - uninitialized errno at exit from -Z /dev/scdN=image.iso;
 * 4.2:
 * - don't print initial bogus progress indicator values;
 * - apparently some firmwares exhibit ambiguity in DVD+R disk
 *   finalizing code;
 * 5.0:
 * - enforced 32K write strategy (needed for DVD-R[W]);
 * - support for DVD-RW Restricted Overwrite Mode;
 * - support for DVD-R[W] Sequential Mode;
 * 5.1:
 * - support for writing speed control;
 * 5.2:
 * - re-make it work under Linux 2.2 kernel;
 * - progress indicator to display recording velocity;
 * - code to protect against overburns;
 * - undocumented -use-the-force-luke flag to overwrite the media
 *   none interactively;
 * - brown-bag bug in "LONG WRITE IN PROGRESS" handling code fixed;
 * 5.3:
 * - Pioneer workarounds/fix-ups, most notably DVR-x05 doesn't seem
 *   to digest immediate "SYNC CACHE";
 * - support for DVD-RW Quick Format, upon release verified to work
 *   with Pioneer DVR-x05;
 * - bug in DVD+RW overburn "protection" code fixed;
 * - media reload is moved here from dvd+rw-format;
 * - refuse to burn if session starts close to or beyond 4GB limit
 *   (limitation of Linux isofs implementation);
 * - dry_run check is postponed all the way till the first write;
 * 5.4:
 * - split first write to two to avoid "empty DMA table?" in kernel log;
 * - setup_fds is introduced to assist ports to another platforms;
 * - set-root-uid assistant code directly at entry point (see main());
 * - OpenBSD/NetBSD support added, it's worth noting that unlike 3.3
 *   port by Maarten Hofman, it's /dev/rcd?c which is expected to be
 *   passed as argument, not /dev/cd?c.
 * 5.5:
 * - fix for ENOENT at unmount, I should have called myself with execlp,
 *   not execl;
 * - security: chdir("/") in set-root-uid assistant;
 * - use /proc/mounts instead of MOUNTED (a.k.a. /etc/mtab) in Linux
 *   umount code;
 */

#define _LARGEFILE_SOURCE 
#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64
#if defined(__linux)
/* ... and "engage" glibc large file support */
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#elif defined(__OpenBSD__) || defined(__NetBSD__)
#define off64_t  off_t
#define stat64   stat
#define fstat64  fstat
#define open64   open
#define pread64  pread
#define pwrite64 pwrite
#endif

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <signal.h>
#include <poll.h>
#include <assert.h>
#include <sys/time.h>

/*
 * Symbols from growisofs_mmc.cpp
 */
/* These never terminate the program, pay attention to the return value	*/
ssize_t poor_mans_pwrite64 (int fd,const void *buff,size_t size,off64_t foff);
int  flush_cache	(int fd);
int  media_reload       (char *,struct stat *);
int  fumount            (int fd);
off64_t get_capacity	(int fd);
int  plus_rw_restart_format(int fd);
int  minus_rw_quickgrow	(int fd);
/* These might terminate the program if error appears fatal, the return	*/
/* value is therefore is always sane					*/
int  get_mmc_profile	(int fd);
int  plusminus_r_C_parm	(int fd,char *C_parm);
int  minus_rw_setup	(int fd);
/* These are atexit/signal handlers					*/
void plus_rw_finalize	();
void plus_r_finalize	();
void minus_rw_finalize	();
void minus_r_finalize	();

/* simplified */
struct iso_primary_descriptor {
    char type	[1];
    char id	[5];
    char void1	[80-5-1];
    char volume_space_size [8];
    char void2	[2048-80-8];
};

#define CD_BLOCK	((off64_t)2048)
#define VOLDESC_OFF	16
#define DVD_BLOCK	(32*1024)

static unsigned int from_733 (unsigned char *s)
{ unsigned int ret=0;
    ret |= s[0];
    ret |= s[1]<<8;
    ret |= s[2]<<16;
    ret |= s[3]<<24;
  return ret;
}

static void to_733 (unsigned char *s,unsigned int val)
{   s[0] = (val)     & 0xFF;
    s[1] = (val>>8)  & 0xFF;
    s[2] = (val>>16) & 0xFF;
    s[3] = (val>>24) & 0xFF;
    s[4] = (val>>24) & 0xFF;
    s[5] = (val>>16) & 0xFF;
    s[6] = (val>>8)  & 0xFF;
    s[7] = (val)     & 0xFF;
}

static ssize_t (*pwrite64_method) (int,const void *,size_t,off64_t) = pwrite64;
/*
 * Page-aligned 2*DVD_BLOCK=64KB large buffer, second 32KB contain
 * "next" session volume descriptors that were written in the beginning
 * of current recording...
 */
static char *the_buffer;
/*
 * id_fd is passed to mkisofs, out_fd - to pwrite and ioctl_fd - to ioctl.
 */
int    in_fd=-1,out_fd=-1,ioctl_fd=-1;

int    dvd_compat=0, speed_factor=0, poor_man=0,
       mmc_profile=-1, quiet=0, overburn=0, no_tty_check=0,
       do_reload=0, dry_run=0;
char  *ioctl_device;
int    _argc;
char **_argv;

#if defined(__linux)

#include <linux/cdrom.h>
#include <sys/ioctl.h>
#include <linux/raw.h>

char *find_raw_device(struct stat64 *sb)
{ int   i,rawctl,dev_major,dev_minor;
  char  *ret=NULL;
  struct raw_config_request req;
  static char rawdevname[16] = "";

    if (!S_ISBLK(sb->st_mode)) return NULL;

    dev_major = major (sb->st_rdev);
    dev_minor = minor (sb->st_rdev);

    if ((rawctl=open ("/dev/rawctl",O_RDONLY)) < 0) return NULL;

    for (i=1;i<256;i++)
    {	req.raw_minor = i;
        if (ioctl(rawctl,RAW_GETBIND,&req) < 0) break;
	if (req.block_major == dev_major && req.block_minor == dev_minor)
	{   sprintf (rawdevname,"/dev/raw/raw%d",i);
	    ret = rawdevname;
	    break;
	}
    }
    close (rawctl);

  return ret;
}

char *setup_fds (char *device)
{ char *odevice;
  uid_t uid=getuid();
  struct stat64 sb;

    /*
     * I ignore return values from set{re}uid calls because if
     * they fail we have no privileges to care about and should
     * just proceed anyway...
     */
    /*
     * Get real, but preserve saved uid. I count that user is
     * logged on console and therefore owns the 'device' [this
     * is a PAM's resposibility to arrange].
     */
    setreuid ((uid_t)-1,uid);

    if ((in_fd = open64 (device,O_RDONLY)) < 0)
	fprintf (stderr,":-( unable to open64(\"%s\",O_RDONLY): ",device),
	perror (NULL), exit (1);

    if (fstat64(in_fd,&sb) < 0)
	fprintf (stderr,":-( unable to stat64(\"%s\"): ",device),
	perror (NULL), exit (1);

    if (!S_ISBLK(sb.st_mode))
    { open_rw:
	if ((out_fd = open64 (device,O_RDWR)) < 0)
	    fprintf (stderr,":-( unable to open64(\"%s\",O_RDWR): ",device),
	    perror (NULL), exit (1);
      opened_rw:
	poor_man = 0;
#if 1
	if (ioctl_fd>=0) close (ioctl_fd), ioctl_fd = -1;
#endif
	setuid (uid);		/* drop all privileges */
	return device;
    }

    ioctl_fd=dup(in_fd);

    /* Consume media_changed flag */
    if (ioctl (ioctl_fd,CDROM_MEDIA_CHANGED,CDSL_CURRENT) < 0)
    {	if (errno == EIO) /* non-CD driver? */ goto open_rw;
	fprintf (stderr,":-( %s: CD_ROM_MEDIA_CHANGED ioctl failed: ",
			device),
	perror (NULL), exit (1);
    }

    mmc_profile = get_mmc_profile (ioctl_fd);
    switch (mmc_profile)
    {	case 0x11:	/* DVD-R			*/
	case 0x13:	/* DVD-RW Restricted Overwrite	*/
	case 0x14:	/* DVD-RW Sequential		*/
	case 0x1B:	/* DVD+R			*/
	open_poor_man:
	    poor_man = 1;
	    out_fd = dup(in_fd);
	    setuid (uid);	/* drop all privileges	*/
	    return device;
	case 0x1A:	/* DVD+RW			*/
	    if (poor_man) goto open_poor_man;
	case 0x12:	/* DVD-RAM			*/
	    break;
	default:
	    fprintf (stderr,":-( %s: unsupported MMC profile %X\n",
			    device,mmc_profile);
	    exit (1);
    }

    /*
     * Attempt to locate /dev/raw/raw*
     */
    setreuid((uid_t)-1,0);	/* get root for a /dev/raw sake */
    if ((odevice=find_raw_device (&sb)))	/* /dev/raw */
    {	if ((out_fd=open64 (odevice,O_RDWR)) < 0)
	{   if (errno == EROFS)	/* must be unpatched kernel */
		goto open_poor_man;
	    else
		fprintf (stderr,":-( unable to open64(\"%s\",O_RDWR): ",
				odevice),
		perror (NULL), exit (1);
	}
	device=odevice;
	goto opened_rw;
    }
    if (mmc_profile == 0x12)	goto open_rw;	/* DVD-RAM */
    else			goto open_poor_man;
}

#elif defined(__OpenBSD__) || defined(__NetBSD__)

char *setup_fds (char *device)
{ uid_t uid=getuid();
  struct stat sb;

    /*
     * We might be entering as euid=root!
     */
    if ((in_fd = open (device,O_RDONLY)) < 0)
	fprintf (stderr,":-( unable to open(\"%s\",O_RDONLY): ",device),
	perror (NULL), exit (1);

    if (fstat (in_fd,&sb) < 0)
	fprintf (stderr,":-( unable to stat(\"%s\"): ",device),
	perror (NULL), exit (1);

    if (!S_ISCHR(sb.st_mode))
    {	if (S_ISBLK(sb.st_mode) && !strncmp (device,"/dev/cd",7))
	{   fprintf (stderr,":-) you most likely want to use /dev/r%s instead!\n",
			    device+5);
	    if (isatty(0) && !dry_run) poll(NULL,0,5000);
	}
	setuid(uid);	/* drop all privileges	*/
	close (in_fd);	/* reopen as mortal	*/
	if ((in_fd = open (device,O_RDONLY)) < 0)
	    fprintf (stderr,":-( unable to reopen(\"%s\",O_RDONLY): ",device),
	    perror (NULL), exit (1);
      open_rw:
	if ((out_fd = open (device,O_RDWR)) < 0)
	    fprintf (stderr,":-( unable to open(\"%s\",O_RDWR): ",device),
	    perror (NULL), exit (1);
      opened_rw:
	poor_man = 0;
	close (ioctl_fd);
	ioctl_fd = -1;
	setuid (uid);	/* drop all privileges	*/
	return device;
    }

    /*
     * Still as euid=root! But note that get_mmc_profile makes sure it's
     * an MMC device, as it terminates the program if the unit doesn't
     * reply to GET CONFIGURATON. In combination with following switch
     * this means that if installed set-root-uid, growisofs grants
     * access to DVD burner(s), but not to any other device.
     */
    ioctl_fd = open (device,O_RDWR); /* O_RDWR is a must for SCIOCCOMMAND */
    if (ioctl_fd < 0)
    {	fprintf (stderr,":-( unable to open(\"%s\",O_RDWR): ",device),
	perror (NULL);
	if (errno == EBUSY)
	    fprintf (stderr,"    is its block counterpart mounted?\n");
	exit (1);
    }

    mmc_profile = get_mmc_profile (ioctl_fd);
    switch (mmc_profile)
    {	case 0x11:	/* DVD-R			*/
	case 0x13:	/* DVD-RW Restricted Overwrite	*/
	case 0x14:	/* DVD-RW Sequential		*/
	case 0x1B:	/* DVD+R			*/
	case 0x1A:	/* DVD+RW			*/
	    poor_man = 1;
	    out_fd = dup(ioctl_fd);
	    setuid (uid);	/* drop all privileges	*/
	    return device;
	case 0x12:	/* DVD-RAM			*/
	    out_fd = dup(ioctl_fd);
	    goto opened_rw;
	default:
	    fprintf (stderr,":-( %s: unsupported MMC profile %X\n",
			    device,mmc_profile);
	    exit (1);
    }
    /* not actually reached */
    setuid(uid);
    goto open_rw;
}

#endif

int setup_C_parm (char *C_parm,struct iso_primary_descriptor *descr)
{ unsigned int next_session=-1;

    if (!poor_man || mmc_profile==0x1A || mmc_profile==0x13)
    {	next_session = from_733(descr->volume_space_size);
	/* pad to closest 32K boundary */
	next_session += 15;
	next_session /= 16;
	next_session *= 16;
	sprintf (C_parm,"16,%u",next_session);
    }
    else if (mmc_profile==0x1B || mmc_profile==0x11 || mmc_profile==0x14)
	next_session=plusminus_r_C_parm (ioctl_fd,C_parm);

  return next_session;
}

int builtin_dd (int infd,int outfd,off64_t outoff)
{ char		*s;
  int   	 n,off,fd;
  struct stat64  sb;
  off64_t	 capacity=0;
  pid_t		 pid=(pid_t)-1;
  volatile struct { time_t zero; off64_t current,final; } *progress;

    if ((fd=mkstemp (s=strdup("/tmp/dvd+rw.XXXXXX"))) < 0)
	fprintf (stderr,":-( unable to mkstemp(\"%s\")",s),
	exit(1);

    ftruncate(fd,sizeof(*progress));
    unlink(s), free(s);

    progress = mmap (NULL,sizeof(*progress),PROT_READ|PROT_WRITE,
			MAP_SHARED,fd,0);
    close (fd);
    if (progress == MAP_FAILED)
	fprintf (stderr,":-( unable to anonymously mmap %ld?\n",
			sizeof(*progress)),
	perror (NULL), exit(1);

    if (fstat64 (infd,&sb))
	perror (":-( unable to fstat64"), exit(1);

    if (ioctl_fd>=0) capacity = get_capacity (ioctl_fd);

    progress->zero=0;
    progress->final=0;
    progress->current=outoff;
    if (S_ISREG(sb.st_mode))
    {	progress->final=sb.st_size;
	if (capacity && progress->final > capacity)
	{   fprintf (stderr,":-( %s: not enough space available "
			    "to accomodate this volume!\n",
			    ioctl_device);
	    if (overburn)
		fprintf (stderr,":-! ignoring...\n");
	    else
		exit (1);
	}
    }
    else if (S_ISCHR(sb.st_mode) || S_ISBLK(sb.st_mode))
	progress->final=capacity;

    if (!dry_run && !quiet && progress->final != 0 && (pid=fork()) == 0)
    { double		ratio,velocity,slept;
      int		delta;
      off64_t		lastcurrent=0,current;
      struct timeval	tv;

	close (infd);
	close (outfd);
	while (1)
	{   gettimeofday (&tv,NULL); slept =  tv.tv_usec/1e6+tv.tv_sec;
	    poll (NULL,0,3333);
	    gettimeofday (&tv,NULL); slept -= tv.tv_usec/1e6+tv.tv_sec;
	    if ((current = progress->current))
	    {	delta = time (NULL) - progress->zero;
		ratio = (double)progress->final/(double)current;
		delta *= ratio - 1.0;
		velocity=(current-lastcurrent)/(-slept*1024*1385);
		fprintf (stdout,"%10lld/%lld (%4.1f%%) @%.1fx, "
				"remaining %d:%02d\n",
				current,progress->final,100.0/ratio,
				velocity,delta/60,delta%60);
		lastcurrent=current;
	    }
	    else
		fprintf (stdout,"%10lld/%lld (%4.1f%%) @0x, remaining ??:??\n",
				(off64_t)0,progress->final,0.0);
	    fflush (stdout);
	}
	exit(0); /* not reached */
    }

    progress->zero=time(NULL);

    /* suck in first 64K and examine ISO9660 Primary Descriptor if present */
    off = 0;
    while ((n=read (infd,the_buffer+off,2*DVD_BLOCK-off)) > 0)
    {	off += n;
	if (off == 2*DVD_BLOCK)
	{   if (!memcmp(the_buffer+DVD_BLOCK,"\1CD001",6))
	    { struct iso_primary_descriptor *descr;
	      off64_t added_size;

		descr=(struct iso_primary_descriptor *)(the_buffer+DVD_BLOCK);
	        added_size=from_733(descr->volume_space_size)*CD_BLOCK;
		if (capacity)
		{   if (outoff+added_size > capacity)
		    {	fprintf (stderr,":-( %s: %lld blocks are free, "
					"%lld to be written\n",
					ioctl_device,
					(capacity-outoff)/2048,
					added_size/2048);
			if (overburn)
			    fprintf (stderr,":-! ignoring...\n");
			else
			{   errno = ENOSPC;
			    n = -1;
			    goto out;
			}
		    }
		}
		else
		    progress->final=outoff+added_size;
	    }

	    if (dry_run) close(infd), close(outfd), exit(0);

	    if (mmc_profile==0x13)	/* Restricted Overwrite mode */
		/*
		 * I can't make it earlier, e.g. upon page05 setup,
		 * as mkisofs wouldn't be able to read otherwise...
		 * How many restrictions does this mode have anyway?
		 */
		minus_rw_quickgrow (ioctl_fd);
#if 0
	    if ((n=(*pwrite64_method) (outfd,the_buffer,2*DVD_BLOCK,outoff))
			!= 2*DVD_BLOCK)
	    {	if (n>0)	errno = EIO;
		else if (n==0)	errno = ENOSPC;
		n = -1;
		goto out;
	    }
	    outoff += 2*DVD_BLOCK;
	    progress->current=outoff;
#else
	    if ((n=(*pwrite64_method) (outfd,the_buffer,DVD_BLOCK,outoff))
			!= DVD_BLOCK)
	    {	if (n>0)	errno = EIO;
		else if (n==0)	errno = ENOSPC;
		n = -1;
		goto out;
	    }
	    outoff += DVD_BLOCK;
	    progress->current=outoff;

	    if ((n=(*pwrite64_method) (outfd,the_buffer+DVD_BLOCK,DVD_BLOCK,outoff))
			!= DVD_BLOCK)
	    {	if (n>0)	errno = EIO;
		else if (n==0)	errno = ENOSPC;
		n = -1;
		goto out;
	    }
	    outoff += DVD_BLOCK;
	    progress->current=outoff;
#endif
	    break;
	}
    }
    if (n<=0) goto out;

    off = 0;
    /*
     * From now on only the first 32K of the_buffer are used!
     */
    while ((n=read (infd,the_buffer+off,DVD_BLOCK-off)) > 0)
    {	off += n;
	if (off == DVD_BLOCK)
	{   off = 0;
	    if ((n=(*pwrite64_method) (outfd,the_buffer,DVD_BLOCK,outoff))
			!= DVD_BLOCK)
	    {	if (n>0)	errno = EIO;
		else if (n==0)	errno = ENOSPC;
		n = -1;
		break;
	    }
	    outoff += DVD_BLOCK;
	    progress->current=outoff;
	}
    }
    if (off)	/* pad to the closest 32K is needed */
    {	memset (the_buffer+off,0,DVD_BLOCK-off);
	if ((n=(*pwrite64_method) (outfd,the_buffer,DVD_BLOCK,outoff))
			!= DVD_BLOCK)
	{    if (n>0)		errno = EIO;
	     else if (n==0)	errno = ENOSPC;
	     n = -1;
	}
	outoff += DVD_BLOCK;
    }
out:
    if (pid!=(pid_t)-1) kill (pid,SIGTERM), waitpid (pid,NULL,0);

  return n;
}

void pipe_mkisofs_up (char *mkisofs_argv[],int infd,int outfd,off64_t outoff)
{ pid_t mkisofs_pid;
  int fildes[2],ret,n;

    if (pipe (fildes) < 0)
        perror (":-( unable to create pipe"), exit(1);

    if ((mkisofs_pid=fork ()) == (pid_t)-1)
	perror (":-( unable to fork mkisofs"), exit(1);
    else if (mkisofs_pid == 0)
    {	dup2  (fildes[1],1);
	close (fildes[0]);
	close (fildes[1]);
	close (outfd);	/* redundant:-) */
	if ((n=fcntl (infd,F_GETFD))<0) n=0;
	fcntl (infd,F_SETFD,n&~FD_CLOEXEC);
	/*
	 * If platform-specific setup_fds did not drop privileges,
	 * do it now. I ignore return value because if it fails,
	 * then privileges were dropped already.
	 */
	setuid(getuid());
	execvp ("mkisofs",mkisofs_argv);
	perror (":-( unable to execute mkisofs"), exit (1);
    }

    close (fildes[1]);

    n=builtin_dd(fildes[0],outfd,outoff);

    if (n==0) /* mkisofs must have finished, consume the exit code */
    {	if ((waitpid (mkisofs_pid,&ret,0)) == -1)
	    perror (":-( waitpid failed"), exit (1);

	if (!WIFEXITED(ret) || WEXITSTATUS(ret)!=0)
	    fprintf (stderr,":-( mkisofs failed: %d\n",errno=ret), exit (1);
    }
    else if (n<0)
    {	perror (":-( write failed"), exit (1);   }
}

/*
 * This may not be larger than 32KB!
 */
#define MAX_IVDs	16
#define IVDs_SIZE	(sizeof(struct iso_primary_descriptor)*MAX_IVDs)

int main (int argc, char *argv[])
{ int  fd,imgfd=-1;
  char *in_device=NULL,*out_device=NULL,*in_image=NULL;
  char dev_found;
  int  i,warn_for_isofs=0;
  char **mkisofs_argv,C_parm[24],M_parm[16];
  int  mkisofs_argc;
  unsigned int next_session=-1,new_size;
  struct iso_primary_descriptor *descr;

    /*
     * This is a set-root-uid "entry point" for listed operations. User
     * can't trick this code to unmount arbitrary file system, as [s]he
     * has to pass opened file descriptor to the mounted device. As for
     * file descriptor passed by this program itself, I rely upon the
     * fact that it was appropriately audited at open time in platform-
     * specific setup_fds above...
     */
    if (*argv[0] == '-')
    { int fd;
      struct stat fdst;

	chdir ("/");
	if (argc != 3)	exit (1);
	fd=atoi (argv[1]);
	if (!strcmp(argv[0],"-umount"))
	{   if (fumount (fd)) exit (errno);
	    exit (0);
	}
	else if (!strcmp(argv[0],"-reload"))
	{   if (fstat (fd,&fdst) < 0)
		perror (":-( unable to fstat"),	exit (errno);
	    close (fd);

	    errno=0;
	    if (media_reload (argv[2],&fdst))
		perror (":-( unable to reload tray");
	    exit (errno);
	}
    }

    /* mmap buffer so that we can use it with /dev/raw/rawN */
    if ((fd=open ("/dev/zero",O_RDWR)) < 0)
	perror (":-( unable to open \"/dev/zero\"?"), exit(1);
    the_buffer = mmap (NULL,2*DVD_BLOCK,PROT_READ|PROT_WRITE,
			MAP_PRIVATE,fd,0);
    close (fd);
    if (the_buffer == MAP_FAILED)
	fprintf (stderr,":-( unable to anonymously mmap %d: ",2*DVD_BLOCK),
	perror (NULL), exit (1);

    mkisofs_argv = malloc ((argc+3)*sizeof(char *));
    if (mkisofs_argv == NULL)
        fprintf (stderr,":-( unable to allocate %ld bytes: ",
			(argc+3)*sizeof(char *)),
	perror (NULL), exit (1);

    mkisofs_argv[0] = "mkisofs", mkisofs_argc = 1;

    _argc=argc,	_argv=argv;

    for (i=1;i<argc;i++)
    { int len=strlen(argv[i]);
	dev_found = '\0';
	if (argv[i][0] == '-')
	{   if (argv[i][1] == 'M')
	    {	if (len > 2) in_device = argv[i]+2;
		else	     in_device = argv[++i];
		dev_found = 'M';
	    }
	    else if (!strncmp(argv[i],"-prev-session",13))
	    {	if (len > 13) in_device = argv[i]+13;
		else          in_device = argv[++i];
		dev_found = 'M';
	    }
	    else if (argv[i][1] == 'Z')
	    {	if (len > 2) in_device = argv[i]+2;
		else	     in_device = argv[++i];
		dev_found = 'Z';
	    }
	    else if (!strncmp(argv[i],"-zero-session",13))
	    {	if (len > 13) in_device = argv[i]+13;
		else          in_device = argv[++i];
		dev_found = 'Z';
	    }
	    else if (!strcmp(argv[i],"-poor-man"))
	    {	poor_man = 1;
		continue;
	    }
	    else if (!strncmp(argv[i],"-speed",6))
	    { char *s;
		if (len>6)	(s=strchr(argv[i],'='))?s++:s;
		else		s=argv[++i];
		if (s)		speed_factor=atoi(s);
		if (speed_factor<=0)
		    fprintf (stderr,"-speed=%d: insane speed factor.\n",
				    speed_factor),
		    exit(1);
		continue;
	    }
	    else if (!strcmp(argv[i],"-dvd-compat"))
	    {	poor_man = 1, dvd_compat = 1;
		continue;
	    }
	    else if (!strcmp(argv[i],"-overburn"))
	    {	overburn = 1;
		continue;
	    }
	    else if (!strcmp(argv[i],"-use-the-force-luke"))
	    {	no_tty_check = 1;
		continue;
	    }
	    else if (!strcmp(argv[i],"-dvd-video"))
		dvd_compat = 1, poor_man = 1;
	    else if (!strcmp(argv[i],"-quiet"))
		quiet = 1;
	    else if (argv[i][1] == 'C' || !strncmp(argv[i],"-cdrecord-params",16))
		continue;
	    else if (argv[i][1] == '#' || !strcmp(argv[i],"-dry-run"))
	    {	dry_run = 1;
		continue;
	    }
	    else if (argv[i][1] == '?' ||
	    		!strcmp(argv[i],"-help") ||
	    		!strcmp(argv[i],"--help")
		    )
	    {	printf ("Usage: %s [-dvd-compat] [-overburn] [-speed=1] \\\n"
			"       -[ZM] /dev/cdrom <mkisofs options>\n",argv[0]);
		printf ("For <mkisofs options> see 'mkisofs %s'\n",argv[i]);
		exit (1);
	    }
	    else if (strstr (argv[i],"-version"))
	    {	printf ("growisofs 5.5, front-ending to ");
		fflush (stdout);
		execlp ("mkisofs","mkisofs","-version",NULL);
		perror (":-( unable to execute mkisofs");
		exit (1);
	    }
	}

	if (dev_found && in_device)
	{   if (*in_device == '=') in_device++;

	    if (dev_found == 'Z')
	    {	if ((in_image = strchr(in_device,'=')))
		{ uid_t euid=geteuid();

		    seteuid (getuid());	/* get real for parsing -Z a=b */

		    while (in_image)
		    {	*in_image='\0';
			errno=0;
			if (access (in_device,F_OK)==0 || errno!=ENOENT)
			    break;
			*in_image='=',
			in_image=strchr(in_image+1,'=');
		    }

		    if (errno)
			fprintf (stderr,":-( \"%s\": unexpected errno:",
					in_device),
			perror (NULL), exit (1);

		    if (in_image)
		    {	in_image++;
			if ((imgfd = open64(in_image,O_RDONLY)) < 0)
			    fprintf (stderr,":-( unable to open64(\"%s\",O_RDONLY): ",
					    in_image),
			    perror (NULL), exit(1);
		    }

		    seteuid (euid);	/* revert to saved [set-]uid */
		}
	    }

	    /*
	     * Sets up in_fd, out_fd, ioctl_fd and poor_man variable.
	     * This procedure is platform-specific. If the program
	     * has to be installed set-root-uid, then this procedure
	     * is the one to drop privileges [if appropriate].
	     */
	    out_device=setup_fds (in_device);

	    *(long *)the_buffer=0;	/* redundant:-) */
            if (pread64 (in_fd,the_buffer,2048,VOLDESC_OFF*CD_BLOCK) != 2048
		&& dev_found == 'M')
	        perror (":-( unable to pread64(2) primary volume descriptor"),
		fprintf (stderr,"    you most likely want to use -Z option.\n"), 
		exit (1);

	    if (dev_found == 'M')
            {	if (memcmp (the_buffer,"\1CD001",6))
	            fprintf (stderr,":-( %s doesn't look like isofs...\n",
		    		in_device), exit(1);

		next_session=setup_C_parm(C_parm,
				(struct iso_primary_descriptor *)the_buffer);

		if (next_session > (0x200000-0x5000)) /* 4GB/2K - 40MB/2K */
		    fprintf (stderr,":-( next session would cross 4GB boundary, "
				    "aborting...\n"),
		    exit (1);

		mkisofs_argv[mkisofs_argc++] = "-C";
		mkisofs_argv[mkisofs_argc++] = C_parm;
		sprintf (M_parm,"/dev/fd/%d",in_fd);
		mkisofs_argv[mkisofs_argc++] = "-M";
		mkisofs_argv[mkisofs_argc++] = M_parm;
		len = 3 + strlen(C_parm) + 3 + strlen(M_parm);
	    }
	    else
	    {	if (!memcmp (the_buffer,"\1CD001",6))
		    warn_for_isofs = 1;
		next_session = 0;
		continue;
	    }
	}
	else
	{   mkisofs_argv[mkisofs_argc++] = argv[i];   }
    }

    if (in_device == NULL)
        fprintf (stderr,"%s: previous \"session\" device is not specified, "
			"do use -M or -Z option\n",argv[0]),
	exit (1);

    mkisofs_argv[mkisofs_argc] = NULL;

    assert (next_session!=-1);
    assert (in_fd!=-1);
    assert (out_fd!=-1);

    if (warn_for_isofs)
    {	if (no_tty_check || isatty (0))
	    fprintf (stderr,"WARNING: %s already carries isofs!\n",in_device),
	    printf ("About to execute '");
	else
	    fprintf (stderr,"FATAL: %s already carries isofs!\n",in_device),
	    exit(1);
    }
    else
	printf ("Executing '");

    if (imgfd>=0)
        printf ("builtin_dd if=%s of=%s obs=32k",in_image,out_device);
    else
    {   for (i=0;i<mkisofs_argc;i++) printf ("%s ",mkisofs_argv[i]);
        printf ("| builtin_dd of=%s obs=32k seek=%u",
		out_device,next_session/16);
    }
    printf ("'\n");
    fflush (stdout);

    if (warn_for_isofs && isatty(0) && !dry_run)
    {	fprintf (stderr,"Sleeping for 5 sec...\a"),	poll (NULL,0,1000);
	fprintf (stderr,"\b\b\b\b\b\b\b\b4 sec...\a"),	poll (NULL,0,1000);
	fprintf (stderr,"\b\b\b\b\b\b\b\b3 sec...\a"),	poll (NULL,0,1000);
	fprintf (stderr,"\b\b\b\b\b\b\b\b2 sec...\a"),	poll (NULL,0,1000);
	fprintf (stderr,"\b\b\b\b\b\b\b\b1 sec...\a"),	poll (NULL,0,1000);
	fprintf (stderr,"\b\b\b\b\b\b\b\b0 sec...\r");
    }

#define CLOSEONEXEC(fd)	do { int f;		\
	if ((f=fcntl ((fd),F_GETFD)) < 0) f=0;	\
	fcntl ((fd),F_SETFD,f|FD_CLOEXEC);	} while (0)
    CLOSEONEXEC(in_fd);
    CLOSEONEXEC(out_fd);
    CLOSEONEXEC(ioctl_fd);
#undef CLOSEONEXEC

    if (!dry_run && (poor_man || next_session==0))	/* unmount media */
    {	pid_t rpid,pid;
	int   rval;

	if ((pid=fork()) == (pid_t)-1)
	    perror (":-( unable to fork -umount"), exit (1);

	/* pass through set-root-uid if any */
	if (pid == 0)
	{   char str[12];

	    if ((rval=fcntl (in_fd,F_GETFD))<0) rval=0;
	    fcntl (in_fd,F_SETFD,rval&~FD_CLOEXEC);

	    sprintf (str,"%d",in_fd);
	    execlp (argv[0],"-umount",str,in_device,NULL);
	    exit (errno);
	}
	while (1)
	{   rpid = waitpid (pid,&rval,0);
	    if (rpid == (pid_t)-1 || (rpid != pid && (errno=ECHILD)))
	    {	if (errno==EINTR)	continue;
		else			perror (":-( waipid failed"), exit(1);
	    }
	    if (WIFSTOPPED(rval))	continue;
	    errno=0;
	    if (WIFEXITED(rval))	errno=WEXITSTATUS(rval);
	    else			errno=ECHILD;
	    break;
	}
	if (errno)
	    fprintf (stderr,":-( unable to unmount %s: ",in_device),
	    perror (""), exit (1);
    }

    if (poor_man) do
    {	ioctl_device=out_device=in_device;

	if (dry_run) break;

	pwrite64_method = poor_mans_pwrite64;

#define	atsignals(func)		do {			\
	signal (SIGHUP,func);	signal (SIGINT,func);	\
	signal (SIGTERM,func);	signal (SIGPIPE,func);	\
	} while (0)
	switch (mmc_profile)
	{   case 0x1A:	/* DVD+RW			*/
		if (dvd_compat)
		    plus_rw_restart_format (ioctl_fd);
		atexit (plus_rw_finalize);
		atsignals (plus_rw_finalize);
		break;
	    case 0x1B: /* DVD+R				*/
		atexit (plus_r_finalize);
#if 1
		/* I'm not sure if this is appropriate, but I need
		 * to try certain thing out first... */
		atsignals (plus_r_finalize);
#endif
		break;
	    case 0x13:	/* DVD-RW Restricted Overwrite	*/
		if (minus_rw_setup (ioctl_fd))
		    perror (":-( unable to setup write page"), exit (1);
		atexit (minus_rw_finalize);
		atsignals (minus_rw_finalize);
		break;
	    case 0x11:	/* DVD-R Sequential		*/
	    case 0x14:	/* DVD-RW Sequential		*/
		if (minus_rw_setup (ioctl_fd))
		    perror (":-( unable to setup write page"), exit (1);
		atexit (minus_r_finalize);
		atsignals (minus_r_finalize);
		break;
	    default:
		fprintf (stderr,":-( %s: mounted media doesn't appear to be "
	    			"DVD+RW or DVD+R\n",out_device),
		exit(1);
		break;
	}
#undef atsignals
    } while (0);

    if (imgfd>=0)
    {   if (builtin_dd (imgfd,out_fd,next_session*CD_BLOCK) < 0)
	    perror (":-( write failed"), exit (1);
	do_reload = 1;
	errno = 0;
	exit (0);
	/* NOT REACHED */
    }

    pipe_mkisofs_up (mkisofs_argv,in_fd,out_fd,next_session*CD_BLOCK);

    if (next_session!=0 && (!poor_man || mmc_profile==0x1A || mmc_profile==0x13))
    {
	if (mmc_profile==0x13)	// DVD-RW Restricted Overwrite
	{   /*
	     * Yet another restriction of Restricted Overwrite mode?
	     * Pioneer DVR-x05 can't read a bit otherwise...
	     */
	    do_reload=0;
	    minus_rw_finalize ();
	}
	else if (ioctl_fd>=0)
	{   /*
	     * (At least) hp100i needs cache flush in order to proceed with
	     * read at this point. This is being moved to kernel, but it won't
	     * ever harm to do this from user-land.
	     */
	    fprintf (stderr,"%s: flushing cache\n",in_device);
	    flush_cache (ioctl_fd);
	}

	/*
	 * recall that second 32KB written in this session are still
	 * in the upper half of the_buffer!
	 */
	descr = (struct iso_primary_descriptor *)the_buffer;

	if (memcmp (descr+16,"\1CD001",6))
	    fprintf (stderr,":-( %s:%d doesn't look like isofs!\n",
				out_device,next_session),
	    errno = EINVAL,
	    exit (1);

	fprintf (stderr,"%s: copying volume descriptor(s)\n",out_device);

	new_size = from_733(descr[16].volume_space_size) + next_session;

	if (pread64 (out_fd,descr,DVD_BLOCK,
		VOLDESC_OFF*CD_BLOCK) != DVD_BLOCK)
	    perror (":-( unable to pread64(2) volume descriptors set"),
	    exit (1);

	memcpy (descr+0,descr+16,sizeof(struct iso_primary_descriptor));
	to_733(descr[0].volume_space_size,new_size);

	/*
	 * copy secondary volume descriptor(s) which are expected to
	 * appear in the very same order.
	 */
	for (i=1;i<MAX_IVDs;i++)
	{   if (descr[16+i].type[0] == (char)255)	break;
	    if (memcmp (descr[16+i].id,"CD001",5))	break;

	    if (descr[16+i].type[0] != descr[i].type[0])
	    {	fprintf (stderr,"volume descriptor mismatch, did you "
				"use same mkisofs options?\n");
		break;
	    }
	    memcpy (descr+i,descr+16+i,sizeof(struct iso_primary_descriptor));
	    to_733(descr[i].volume_space_size,new_size);
	}

	if ((*pwrite64_method) (out_fd,descr,DVD_BLOCK,
		VOLDESC_OFF*CD_BLOCK) != DVD_BLOCK)
            perror (":-( unable to pwrite64(2) volume descriptors set"),
	    exit (1);

	if (ioctl_fd >= 0) flush_cache (ioctl_fd);
    }

    do_reload = 1;
    errno = 0;
    exit (0);
}
